"""A module to handle OME-NGFF images stored in Zarr format."""

from typing import Any, Literal

from ngio.core.image_like_handler import ImageLike
from ngio.core.roi import WorldCooROI
from ngio.io import AccessModeLiteral, StoreOrGroup
from ngio.ngff_meta import PixelSize
from ngio.ngff_meta.fractal_image_meta import ImageMeta
from ngio.utils._common_types import ArrayLike


class Image(ImageLike):
    """A class to handle OME-NGFF images stored in Zarr format.

    This class provides methods to load image data and metadata from
    an OME-Zarr file.
    """

    def __init__(
        self,
        store: StoreOrGroup,
        *,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = False,
        strict: bool = True,
        cache: bool = True,
        mode: AccessModeLiteral = "r+",
        label_group: Any = None,
    ) -> None:
        """Initialize the the Image Object.

        Note: Only one of `path`, `idx`, 'pixel_size' or 'highest_resolution'
        should be provided.

        store (StoreOrGroup): The Zarr store or group containing the image data.
        path (str | None): The path to the level.
        idx (int | None): The index of the level.
        pixel_size (PixelSize | None): The pixel size of the level.
        highest_resolution (bool): Whether to get the highest resolution level.
        strict (bool): Whether to raise an error where a pixel size is not found
            to match the requested "pixel_size".
        cache (bool): Whether to cache the metadata.
        mode (AccessModeLiteral): The mode to open the group in.
        label_group: The group containing the labels.
        """
        super().__init__(
            store=store,
            path=path,
            idx=idx,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            strict=strict,
            meta_mode="image",
            cache=cache,
            mode=mode,
            _label_group=label_group,
        )

    @property
    def metadata(self) -> ImageMeta:
        """Return the metadata of the image."""
        meta = super().metadata
        assert isinstance(meta, ImageMeta)
        return meta

    @property
    def channel_labels(self) -> list[str]:
        """Return the names of the channels in the image."""
        return self.metadata.channel_labels

    def get_channel_idx(
        self,
        label: str | None = None,
        wavelength_id: str | None = None,
    ) -> int:
        """Return the index of the channel."""
        return self.metadata.get_channel_idx(label=label, wavelength_id=wavelength_id)

    def get_array_from_roi(
        self,
        roi: WorldCooROI,
        c: int | str | None = None,
        t: int | slice | None = None,
        mode: Literal["numpy"] | Literal["dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            c (int | str | None): The channel index or label.
            t (int | slice | None): The time index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.

        Returns:
            ArrayLike: The image data.
        """
        if isinstance(c, str):
            c = self.get_channel_idx(label=c)

        return self._get_array_from_roi(
            roi=roi, t=t, c=c, mode=mode, preserve_dimensions=preserve_dimensions
        )

    def set_array_from_roi(
        self,
        patch: ArrayLike,
        roi: WorldCooROI,
        c: int | str | None = None,
        t: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            patch (ArrayLike): The patch to set.
            c (int | str | None): The channel index or label.
            t (int | slice | None): The time index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.

        """
        if isinstance(c, str):
            c = self.get_channel_idx(label=c)

        return self._set_array_from_roi(
            patch=patch, roi=roi, t=t, c=c, preserve_dimensions=preserve_dimensions
        )

    def get_array(
        self,
        *,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        c: int | str | None = None,
        t: int | slice | None = None,
        mode: Literal["numpy"] | Literal["dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the image data.

        Args:
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            c (int | str | None): The channel index or label.
            t (int | slice | None): The time index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.

        Returns:
            ArrayLike: The image data.
        """
        if isinstance(c, str):
            c = self.get_channel_idx(label=c)

        return self._get_array(
            x=x,
            y=y,
            z=z,
            t=t,
            c=c,
            mode=mode,
            preserve_dimensions=preserve_dimensions,
        )

    def set_array(
        self,
        patch: ArrayLike,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        c: int | str | None = None,
        t: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the image data in the zarr array.

        Args:
            patch (ArrayLike): The patch to set.
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            c (int | str | None): The channel index or label.
            t (int | slice | None): The time index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        if isinstance(c, str):
            c = self.get_channel_idx(label=c)
        return self._set_array(
            patch=patch,
            x=x,
            y=y,
            z=z,
            t=t,
            c=c,
            preserve_dimensions=preserve_dimensions,
        )

    def consolidate(self, order: Literal[0, 1, 2] = 1) -> None:
        """Consolidate the image."""
        self._consolidate(order=order)



"""Various utilities for the ngio package."""

import os

from ngio.utils._common_types import ArrayLike
from ngio.utils._errors import (
    NgioFileExistsError,
    NgioFileNotFoundError,
    NgioNGFFValidationError,
    NgioTableValidationError,
)
from ngio.utils._logger import ngio_logger, set_logger_level
from ngio.utils._pydantic_utils import BaseWithExtraFields, unique_items_validator

set_logger_level(os.getenv("NGIO_LOGGER_LEVEL", "WARNING"))

__all__ = [
    "ArrayLike",
    # Pydantic
    "BaseWithExtraFields",
    "unique_items_validator",
    # Logger
    "ngio_logger",
    "set_logger_level",
    # Errors
    "NgioFileExistsError",
    "NgioFileNotFoundError",
    "NgioNGFFValidationError",
    "NgioTableValidationError",
]



# Create a generic error class for the NGFF project


class NgioError(Exception):
    """Base class for all errors in the NGFF project."""

    pass


class NgioFileNotFoundError(NgioError):
    """Error raised when a file is not found."""

    pass


class NgioFileExistsError(NgioError):
    """Error raised when a file already exists."""

    pass


class NgioNGFFValidationError(NgioError):
    """Error raised when a file does not pass validation."""

    pass


class NgioTableValidationError(NgioError):
    """Error raised when a table does not pass validation."""

    pass



from functools import partial
from typing import Literal

import dask.array as da
import numpy as np
import zarr
from scipy.ndimage import zoom


def _zoom_inputs_check(
    source_array: np.ndarray | da.Array,
    scale: tuple[int, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
) -> tuple[np.ndarray, tuple[int, ...]]:
    if scale is None and target_shape is None:
        raise ValueError("Either scale or target_shape must be provided")

    if scale is not None and target_shape is not None:
        raise ValueError("Only one of scale or target_shape must be provided")

    if scale is None:
        assert target_shape is not None, "Target shape must be provided"
        assert len(target_shape) == source_array.ndim, (
            "Target shape must have the "
            "same number of dimensions as "
            "the source array"
        )
        _scale = np.array(target_shape) / np.array(source_array.shape)
        _target_shape = target_shape
    else:
        _scale = np.array(scale)
        _target_shape = tuple(np.array(source_array.shape) * scale)

    return _scale, _target_shape


def _dask_zoom(
    source_array: da.Array,
    scale: tuple[int, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
    order: Literal[0, 1, 2] = 1,
) -> da.Array:
    """Dask implementation of zooming an array.

    Only one of scale or target_shape must be provided.

    Args:
        source_array (da.Array): The source array to zoom.
        scale (tuple[int, ...] | None): The scale factor to zoom by.
        target_shape (tuple[int, ...], None): The target shape to zoom to.
        order (Literal[0, 1, 2]): The order of interpolation. Defaults to 1.

    Returns:
        da.Array: The zoomed array.
    """
    # This function follow the implementation from:
    # https://github.com/ome/ome-zarr-py/blob/master/ome_zarr/dask_utils.py
    # The module was contributed by Andreas Eisenbarth @aeisenbarth
    # See https://github.com/toloudis/ome-zarr-py/pull/

    _scale, _target_shape = _zoom_inputs_check(
        source_array=source_array, scale=scale, target_shape=target_shape
    )

    # Rechunk to better match the scaling operation
    source_chunks = np.array(source_array.chunksize)
    better_source_chunks = np.maximum(1, np.round(source_chunks * _scale) / _scale)
    better_source_chunks = better_source_chunks.astype(int)
    source_array = source_array.rechunk(better_source_chunks)  # type: ignore

    # Calculate the block output shape
    block_output_shape = tuple(np.ceil(better_source_chunks * _scale).astype(int))

    zoom_wrapper = partial(
        zoom, zoom=_scale, order=order, mode="grid-constant", grid_mode=True
    )

    out_array = da.map_blocks(
        zoom_wrapper, source_array, chunks=block_output_shape, dtype=source_array.dtype
    )

    # Slice and rechunk to target
    slices = tuple(slice(0, ts, 1) for ts in _target_shape)
    out_array = out_array[slices]
    return out_array


def _numpy_zoom(
    source_array: np.ndarray,
    scale: tuple[int, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
    order: Literal[0, 1, 2] = 1,
) -> np.ndarray:
    """Numpy implementation of zooming an array.

    Only one of scale or target_shape must be provided.

    Args:
        source_array (np.ndarray): The source array to zoom.
        scale (tuple[int, ...] | None): The scale factor to zoom by.
        target_shape (tuple[int, ...], None): The target shape to zoom to.
        order (Literal[0, 1, 2]): The order of interpolation. Defaults to 1.

    Returns:
        np.ndarray: The zoomed array
    """
    _scale, _ = _zoom_inputs_check(
        source_array=source_array, scale=scale, target_shape=target_shape
    )

    out_array = zoom(
        source_array, zoom=_scale, order=order, mode="grid-constant", grid_mode=True
    )
    assert isinstance(out_array, np.ndarray)
    return out_array


def on_disk_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: Literal[0, 1, 2] = 1,
    mode: Literal["dask", "numpy"] = "dask",
) -> None:
    """Apply a zoom operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to zoom.
        target (zarr.Array): The target array to save the zoomed result to.
        order (Literal[0, 1, 2]): The order of interpolation. Defaults to 1.
        mode (Literal["dask", "numpy"]): The mode to use. Defaults to "dask".
    """
    if not isinstance(source, zarr.Array):
        raise ValueError("source must be a zarr array")

    if not isinstance(target, zarr.Array):
        raise ValueError("target must be a zarr array")

    if source.dtype != target.dtype:
        raise ValueError("source and target must have the same dtype")

    assert mode in ["dask", "numpy"], "mode must be either 'dask' or 'numpy'"

    if mode == "numpy":
        target[...] = _numpy_zoom(source[...], target_shape=target.shape, order=order)

    source_array = da.from_zarr(source)
    target_array = _dask_zoom(source_array, target_shape=target.shape, order=order)

    target_array = target_array.rechunk(target.chunks)
    target_array.to_zarr(target)


def on_disk_coarsen(
    source: zarr.Array,
    target: zarr.Array,
    aggregation_function: np.ufunc,
    coarsening_setup: dict[int, int],
) -> None:
    """Apply a coarsening operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to coarsen.
        target (zarr.Array): The target array to save the coarsened result to.
        aggregation_function (np.ufunc): The aggregation function to use.
        coarsening_setup (dict[int, int]): The coarsening setup to use.
    """
    source_array = da.from_zarr(source)

    for ax, factor in coarsening_setup.items():
        if ax >= source_array.ndim:
            raise ValueError(
                "Coarsening axis must be less than the number of dimensions"
            )
        if factor <= 0:
            raise ValueError("Coarsening factor must be greater than 0")

        assert isinstance(factor, int), "Coarsening factor must be an integer"

    out_target = da.coarsen(
        aggregation_function, source_array, coarsening_setup, trim_excess=True
    )
    out_target = out_target.rechunk(target.chunks)
    out_target.to_zarr(target)



from typing import Protocol

from scipy.ndimage import zoom

from ngio.utils._common_types import ArrayLike


class Transform(Protocol):
    """A protocol for data transforms to be performed on image data."""

    def get(self, data: ArrayLike) -> ArrayLike:
        """Apply the transform to the data and return the result."""
        ...

    def set(self, data: ArrayLike) -> ArrayLike:
        """Apply the reverse transform to the data and return the result."""
        ...


class ZoomTransform:
    """A transform to zoom in or out of the data."""

    def __init__(self, zoom_factor: list[float]):
        """Initialize the ZoomTransform object."""
        self.zoom_factor = zoom_factor

    def get(self, data: ArrayLike) -> ArrayLike:
        """Apply the zoom transform to the data and return the result."""
        return zoom(data, self.zoom_factor)

    def set(self, data: ArrayLike) -> ArrayLike:
        """Apply the reverse zoom transform to the data and return the result."""
        return zoom(data, [1 / factor for factor in self.zoom_factor])
